class BaseTableError(BaseException):
    '''
    Base exception for the table package from which all other exceptions are
    derived
    '''

class TableEmptyError(BaseTableError):
    '''
    Raised when an operation which requires the table be populated with some
    data is attempted on an empty table
    '''



from .row import Row, Header, LineBreak
from .cell import Cell
from .exceptions import TableEmptyError

class Table:

    def __init__(self, *args, **kwargs):
        '''Returns a new Table instance

        Arguments:
            hborder     - Single character used to draw horizontal border
            vborder     - Single character used to draw vertical border
            fgcolor     - Foreground colour to print table in
            bgcolor     - Background colour to print table in
            align       - Where to align text in table
            bold        - Display text in table bold
            dark        - Use a darker shade of the supplied fgcolor
        '''

        self.kwargs = {}
        self.kwargs['hborder'] = kwargs.get('hborder', '-')
        self.kwargs['vborder'] = kwargs.get('vborder', '|')
        self.kwargs['fgcolor'] = kwargs.get('fgcolor', None)
        self.kwargs['bgcolor'] = kwargs.get('bgcolor', None)
        self.kwargs['align']   = kwargs.get('align', 'left')
        self.kwargs['bold']    = kwargs.get('bold', False)
        self.kwargs['dark']    = kwargs.get('dark', False)

        self.rows = []

    def add_row(self, cells, **kwargs):
        '''Add a row to the table

        Arguments:
            cells       - List of Cell objects or strings
            header      - Treat row as a header
            fgcolor     - Foreground color to use when printing text
            bgcolor     - Background color to use when printing text
            align       - Where to align text in the Row
            bold        - Display text in row as bold
            italic      - Display text in row in italics
        '''

        if kwargs.get('header', False) and 'bold' not in kwargs:
            kwargs['bold'] = True
            self.rows.append(Header(self, cells, **kwargs))

        else:
            self.rows.append(Row(self, cells, **kwargs))

    def add_linebreak(self, character, **kwargs):
        '''Add a linebreak to the table

        Arguments:
            character   - Single character to use as a line break
            fgcolor     - Foreground colour to use when printing
            bgcolor     - Background colour to use when printing
            bold        - Display text as bold
            dark        - User a darker shade of the supplied fgcolor
        '''

        self.rows.append(LineBreak(self, [character], **kwargs))

    def column_width(self, column_index):
        '''Return width of given column'''

        cells = []
        for row in self.rows:
            try:
                cells.append(len(row.cells[column_index].text))
            except IndexError:
                cells.append(0)

        return max(cells)

    def as_string(self):
        '''Return table as a string to be printed'''

        if not self.rows:
            raise TableEmptyError('Unable to print empty table')

        output = []
        for row in self.rows:
            output.append(row.as_string())

        return '\n' + '\n'.join(output) + '\n'

    def __str__(self):
        return self.as_string()



from .cell import Cell

class Row:

    def __init__(self, parent, cells, **kwargs):
        '''Returns a new Row instance

        Arguments:
            parent      - Reference to Table
            cells       - List of cell objects or strings
            fgcolor     - Foreground colour to print row in
            bgcolor     - Background colour to print row in
            align       - Where to align text in row
            bold        - Display row text in bold
            dark        - Use a darker shade of the supplied fgcolor
        '''

        self.parent = parent
        self.kwargs = kwargs

        self.cells = []
        for item in cells:
            tmp_kwargs = {**self.parent.kwargs}
            tmp_kwargs.update(kwargs)
            if type(item) is not Cell:
                cell = Cell(item, **tmp_kwargs)
            else:
                tmp_kwargs.update(item.kwargs)
                cell = Cell(item.text, **tmp_kwargs)

            self.cells.append(cell)

    def as_string(self):
        '''Return Row object as a string'''

        row = []
        vborder = Cell(self.parent.kwargs['vborder'], **self.kwargs).as_string(0, '')
        for index, cell in enumerate(self.cells):
            row.append(cell.as_string(self.parent.column_width(index)))

        return ' ' + f'{vborder}'.join(row)

class Header(Row):

    def as_string(self):
        output = []
        border = [self.parent.kwargs['hborder'] for c in self.cells]
        output.append(LineBreak(self.parent, border, **self.kwargs).as_string())
        output.append(Row(self.parent, self.cells, **self.kwargs).as_string())
        output.append(LineBreak(self.parent, border, **self.kwargs).as_string())

        return '\n'.join(output)

class LineBreak(Row):

    def as_string(self):

        columns = 0
        for row in self.parent.rows:
            columns = len(row.cells) if len(row.cells) > columns else columns

        border_char = self.cells[0].text
        row = []
        for column in range(columns):
            row.append(Cell(border_char * self.parent.column_width(column), **self.kwargs).as_string(0, border_char))
            row.append(Cell(border_char, **self.kwargs).as_string(0, ''))

        return ' ' + ''.join(row[:-1]) + ' '



import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='expresstable',
    version='0.1.1',
    author='Christopher Smith',
    description='Quickly create tables in Python',
    url='https://github.com/christopherdavidsmith/expresstable',
    packages=setuptools.find_packages(),
    install_requires=[
        'termcolor>=1.1.0'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ]
)



from termcolor import colored

class Cell:

    def __init__(self, text, **kwargs):
        '''Returns a new Cell instance

        Arguments:
            text        - Text to appear inside this Cell
            fgcolor     - Foreground color to use when printing text
            bgcolor     - Background color to use when printing text
            align       - Where to align text in the Cell
            bold        - Display text in Cell as bold
            dark        - Use a darker shade of the supplied fgcolor
        '''

        self.text = str(text)
        self.kwargs = kwargs

    def as_string(self, padding, padding_character=' '):
        '''Return Cell object as a string'''

        # Prepend 'on_' to bgcolor as it will be passed to termcolor
        if 'bgcolor' in self.kwargs and self.kwargs['bgcolor'] is not None:
            self.kwargs['bgcolor'] = f'on_{self.kwargs["bgcolor"]}'

        # Calculate padding
        padding = ' ' * (padding - len(self.text))
        pre_padding = padding[0:int(len(padding) / 2)]
        post_padding = padding[len(pre_padding):]

        # Generate attributes
        attributes = [
            attribute for attribute in ['bold', 'dark']
            if self.kwargs.get(attribute, None)
        ]

        # Generate text
        if self.kwargs.get('align', None) == 'right':
            text = f'{padding}{self.text}'

        elif self.kwargs.get('align', None) == 'center':
            text = f'{pre_padding}{self.text}{post_padding}'

        else:
            text = f'{self.text}{padding}'

        # Format text
        text = colored(
            f'{padding_character}{text}{padding_character}',
            self.kwargs.get('fgcolor', None),
            self.kwargs.get('bgcolor', None),
            attrs=attributes
        )

        return text
